@@ -12,13 +12,15 @@
 
 """Run a projects tests and load them into stestr."""
 
+import errno
+import functools
+import io
 import os
 import subprocess
 import sys
 import warnings
 
 from cliff import command
-import six
 import subunit
 import testtools
 
@@ -30,6 +32,8 @@ from stestr import output
 from stestr.repository import abstract as repository
 from stestr.repository import util
 from stestr import results
+from stestr.subunit_runner import program
+from stestr.subunit_runner import run as subunit_run
 from stestr.testlist import parse_list
 from stestr import user_config
 
@@ -41,7 +45,7 @@ def _to_int(possible, default=0, out=sys.stderr):
         i = default
         msg = ('Unable to convert "%s" to an integer.  Using %d.\n' %
                (possible, default))
-        out.write(six.text_type(msg))
+        out.write(str(msg))
     return i
 
 
@@ -121,7 +125,14 @@ class Run(command.Command):
                             help="Takes in a single test to bypasses test "
                             "discover and just execute the test specified. A "
                             "file may be used in place of a test name.")
-        parser.add_argument('--random', '-r', action="store_true",
+        parser.add_argument('--pdb', default=None, metavar='TEST_ID',
+                            help="Run a single test id with the intent of "
+                            "using pdb. This does not launch any separate "
+                            "processes to ensure pdb works as expected. It "
+                            "will bypass test discovery and just execute the "
+                            "test specified. A file may be used in place of a "
+                            "test name.")
+        parser.add_argument('--random', action="store_true",
                             default=False,
                             help="Randomize the test order after they are "
                             "partitioned into separate workers")
@@ -156,6 +167,10 @@ class Run(command.Command):
                             dest='all_attachments',
                             help='If set print all text attachment contents on'
                             ' a successful test execution')
+        parser.add_argument('--show-binary-attachments', action='store_true',
+                            dest='show_binary_attachments',
+                            help='If set, show non-text attachments. This is '
+                            'generally only useful for debug purposes.')
         return parser
 
     def take_action(self, parsed_args):
@@ -232,7 +247,9 @@ class Run(command.Command):
             filters=filters, pretty_out=pretty_out, color=color,
             stdout=stdout, abbreviate=abbreviate,
             suppress_attachments=suppress_attachments,
-            all_attachments=all_attachments)
+            all_attachments=all_attachments,
+            show_binary_attachments=args.show_binary_attachments,
+            pdb=args.pdb)
 
         # Always output slowest test info if requested, regardless of other
         # test run options
@@ -273,7 +290,8 @@ def run_command(config='.stestr.conf', repo_type='file',
                 no_discover=False, random=False, combine=False, filters=None,
                 pretty_out=True, color=False, stdout=sys.stdout,
                 abbreviate=False, suppress_attachments=False,
-                all_attachments=False):
+                all_attachments=False, show_binary_attachments=True,
+                pdb=False):
     """Function to execute the run command
 
     This function implements the run command. It will run the tests specified
@@ -336,6 +354,11 @@ def run_command(config='.stestr.conf', repo_type='file',
         will not print attachments on successful test execution.
     :param bool all_attachments: When set true subunit_trace will print all
         text attachments on successful test execution.
+    :param bool show_binary_attachments: When set to true, subunit_trace will
+        print binary attachments in addition to text attachments.
+    :param str pdb: Takes in a single test_id to bypasses test
+        discover and just execute the test specified without launching any
+        additional processes. A file name may be used in place of a test name.
 
     :return return_code: The exit code for the command. 0 for success and > 0
         for failures.
@@ -354,7 +377,17 @@ def run_command(config='.stestr.conf', repo_type='file',
                    "--test-path ")
             stdout.write(msg)
             exit(1)
-        repo = util.get_repo_initialise(repo_type, repo_url)
+        try:
+            repo = util.get_repo_initialise(repo_type, repo_url)
+        except OSError as e:
+            if e.errno != errno.EEXIST:
+                raise
+            repo_path = repo_url or './stestr'
+            stdout.write('The specified repository directory %s already '
+                         'exists. Please check if the repository already '
+                         'exists or select a different path\n' % repo_path)
+            return 1
+
     combine_id = None
     concurrency = _to_int(concurrency)
 
@@ -365,7 +398,18 @@ def run_command(config='.stestr.conf', repo_type='file',
         return 2
     if combine:
         latest_id = repo.latest_id()
-        combine_id = six.text_type(latest_id)
+        combine_id = str(latest_id)
+    if no_discover and pdb:
+        msg = ("--no-discover and --pdb are mutually exclusive options, "
+               "only specify one at a time")
+        stdout.write(msg)
+        return 2
+    if pdb and until_failure:
+        msg = ("pdb mode does not function with the --until-failure flag, "
+               "only specify one at a time")
+        stdout.write(msg)
+        return 2
+
     if no_discover:
         ids = no_discover
         if '::' in ids:
@@ -394,7 +438,8 @@ def run_command(config='.stestr.conf', repo_type='file',
                              pretty_out=pretty_out,
                              color=color, stdout=stdout, abbreviate=abbreviate,
                              suppress_attachments=suppress_attachments,
-                             all_attachments=all_attachments)
+                             all_attachments=all_attachments,
+                             show_binary_attachments=show_binary_attachments)
 
         if not until_failure:
             return run_tests()
@@ -418,6 +463,30 @@ def run_command(config='.stestr.conf', repo_type='file',
                 if result:
                     return result
 
+    if pdb:
+        ids = pdb
+        if '::' in ids:
+            ids = ids.replace('::', '.')
+        if ids.find('/') != -1:
+            root = ids.replace('.py', '')
+            ids = root.replace('/', '.')
+        runner = subunit_run.SubunitTestRunner
+        stream = io.BytesIO()
+        program.TestProgram(module=None, argv=['stestr', ids],
+                            testRunner=functools.partial(runner,
+                                                         stdout=stream))
+        stream.seek(0)
+        run_proc = [('subunit', stream)]
+        return load.load(in_streams=run_proc,
+                         subunit_out=subunit_out,
+                         repo_type=repo_type,
+                         repo_url=repo_url, run_id=combine_id,
+                         pretty_out=pretty_out,
+                         color=color, stdout=stdout, abbreviate=abbreviate,
+                         suppress_attachments=suppress_attachments,
+                         all_attachments=all_attachments,
+                         show_binary_attachments=show_binary_attachments)
+
     if failing or analyze_isolation:
         ids = _find_failing(repo)
     else:
@@ -466,7 +535,8 @@ def run_command(config='.stestr.conf', repo_type='file',
                     repo_type=repo_type, repo_url=repo_url,
                     pretty_out=pretty_out, color=color, abbreviate=abbreviate,
                     stdout=stdout, suppress_attachments=suppress_attachments,
-                    all_attachments=all_attachments)
+                    all_attachments=all_attachments,
+                    show_binary_attachments=show_binary_attachments)
                 if run_result > result:
                     result = run_result
             return result
@@ -481,7 +551,8 @@ def run_command(config='.stestr.conf', repo_type='file',
                               stdout=stdout,
                               abbreviate=abbreviate,
                               suppress_attachments=suppress_attachments,
-                              all_attachments=all_attachments)
+                              all_attachments=all_attachments,
+                              show_binary_attachments=show_binary_attachments)
     else:
         # Where do we source data about the cause of conflicts.
         latest_run = repo.get_latest_run()
@@ -526,7 +597,7 @@ def _run_tests(cmd, until_failure,
                subunit_out=False, combine_id=None, repo_type='file',
                repo_url=None, pretty_out=True, color=False, stdout=sys.stdout,
                abbreviate=False, suppress_attachments=False,
-               all_attachments=False):
+               all_attachments=False, show_binary_attachments=False):
     """Run the tests cmd was parameterised with."""
     cmd.setUp()
     try:
@@ -544,7 +615,8 @@ def _run_tests(cmd, until_failure,
                              pretty_out=pretty_out, color=color, stdout=stdout,
                              abbreviate=abbreviate,
                              suppress_attachments=suppress_attachments,
-                             all_attachments=all_attachments)
+                             all_attachments=all_attachments,
+                             show_binary_attachments=show_binary_attachments)
 
         if not until_failure:
             return run_tests()
